/*
 * importResolver.test.ts
 *
 * importResolver tests.
 */

import assert from 'assert';

import { ImportResolver } from '../analyzer/importResolver';
import { ConfigOptions } from '../common/configOptions';
import { lib, sitePackages, typeshedFallback } from '../common/pathConsts';
import { combinePaths, getDirectoryPath, normalizeSlashes } from '../common/pathUtils';
import { PyrightFileSystem } from '../pyrightFileSystem';
import { TestFileSystem } from './harness/vfs/filesystem';

const libraryRoot = combinePaths(normalizeSlashes('/'), lib, sitePackages);

test('partial stub file exists', () => {
    const files = [
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', 'partialStub.pyi'),
            content: 'def test(): ...',
        },
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', 'py.typed'),
            content: 'partial\n',
        },
        {
            path: combinePaths(libraryRoot, 'myLib', 'partialStub.py'),
            content: 'def test(): pass',
        },
    ];

    const importResult = getImportResult(files, ['myLib', 'partialStub']);
    assert(importResult.isImportFound);
    assert(importResult.isStubFile);
    assert.strictEqual(
        1,
        importResult.resolvedPaths.filter((f) => f === combinePaths(libraryRoot, 'myLib', 'partialStub.pyi')).length
    );
});

test('partial stub __init__ exists', () => {
    const files = [
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', '__init__.pyi'),
            content: 'def test(): ...',
        },
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', 'py.typed'),
            content: 'partial\n',
        },
        {
            path: combinePaths(libraryRoot, 'myLib', '__init__.py'),
            content: 'def test(): pass',
        },
    ];

    const importResult = getImportResult(files, ['myLib']);
    assert(importResult.isImportFound);
    assert(importResult.isStubFile);
    assert.strictEqual(
        1,
        importResult.resolvedPaths.filter((f) => f === combinePaths(libraryRoot, 'myLib', '__init__.pyi')).length
    );
});

test('side by side files', () => {
    const myFile = combinePaths('src', 'file.py');
    const files = [
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', 'partialStub.pyi'),
            content: 'def test(): ...',
        },
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', 'py.typed'),
            content: 'partial\n',
        },
        {
            path: combinePaths(libraryRoot, 'myLib', 'partialStub.pyi'),
            content: '# empty',
        },
        {
            path: combinePaths(libraryRoot, 'myLib', 'partialStub.py'),
            content: 'def test(): pass',
        },
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', 'partialStub2.pyi'),
            content: 'def test(): ...',
        },
        {
            path: combinePaths(libraryRoot, 'myLib', 'partialStub2.py'),
            content: 'def test(): pass',
        },
        {
            path: myFile,
            content: '# not used',
        },
    ];

    const fs = createFileSystem(files);
    const configOptions = getConfigOption(fs);
    const importResolver = new ImportResolver(fs, configOptions);

    // Real side by side stub file win over virtual one.
    const sideBySideResult = importResolver.resolveImport(myFile, configOptions.findExecEnvironment(myFile), {
        leadingDots: 0,
        nameParts: ['myLib', 'partialStub'],
        importedSymbols: [],
    });

    assert(sideBySideResult.isImportFound);
    assert(sideBySideResult.isStubFile);

    const sideBySideStubFile = combinePaths(libraryRoot, 'myLib', 'partialStub.pyi');
    assert.strictEqual(1, sideBySideResult.resolvedPaths.filter((f) => f === sideBySideStubFile).length);
    assert.strictEqual('# empty', fs.readFileSync(sideBySideStubFile, 'utf8'));

    // Side by side stub doesn't completely disable partial stub.
    const partialStubResult = importResolver.resolveImport(myFile, configOptions.findExecEnvironment(myFile), {
        leadingDots: 0,
        nameParts: ['myLib', 'partialStub2'],
        importedSymbols: [],
    });

    assert(partialStubResult.isImportFound);
    assert(partialStubResult.isStubFile);

    const partialStubFile = combinePaths(libraryRoot, 'myLib', 'partialStub2.pyi');
    assert.strictEqual(1, partialStubResult.resolvedPaths.filter((f) => f === partialStubFile).length);
});

test('stub package', () => {
    const files = [
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', 'stub.pyi'),
            content: '# empty',
        },
        {
            path: combinePaths(libraryRoot, 'myLib', 'partialStub.py'),
            content: 'def test(): pass',
        },
    ];

    // If fully typed stub package exists, that wins over the real package.
    const importResult = getImportResult(files, ['myLib', 'partialStub']);
    assert(!importResult.isImportFound);
});

test('stub in typing folder over partial stub package', () => {
    const typingFolder = combinePaths(normalizeSlashes('/'), 'typing');
    const files = [
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', '__init__.pyi'),
            content: 'def test(): ...',
        },
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', 'py.typed'),
            content: 'partial\n',
        },
        {
            path: combinePaths(typingFolder, 'myLib.pyi'),
            content: '# empty',
        },
        {
            path: combinePaths(libraryRoot, 'myLib', '__init__.py'),
            content: 'def test(): pass',
        },
    ];

    // If the package exists in typing folder, that gets picked up first.
    const importResult = getImportResult(files, ['myLib'], (c) => (c.stubPath = typingFolder));
    assert(importResult.isImportFound);
    assert(importResult.isStubFile);
    assert.strictEqual(
        0,
        importResult.resolvedPaths.filter((f) => f === combinePaths(libraryRoot, 'myLib', '__init__.pyi')).length
    );
});

test('partial stub package in typing folder', () => {
    const typingFolder = combinePaths(normalizeSlashes('/'), 'typing');
    const files = [
        {
            path: combinePaths(typingFolder, 'myLib-stubs', '__init__.pyi'),
            content: 'def test(): ...',
        },
        {
            path: combinePaths(typingFolder, 'myLib-stubs', 'py.typed'),
            content: 'partial\n',
        },
        {
            path: combinePaths(libraryRoot, 'myLib', '__init__.py'),
            content: 'def test(): pass',
        },
    ];

    const importResult = getImportResult(files, ['myLib'], (c) => (c.stubPath = typingFolder));
    assert(importResult.isImportFound);
    assert(importResult.isStubFile);
    assert.strictEqual(
        1,
        importResult.resolvedPaths.filter((f) => f === combinePaths(libraryRoot, 'myLib', '__init__.pyi')).length
    );
});

test('typeshed folder', () => {
    const typeshedFolder = combinePaths(normalizeSlashes('/'), 'ts');
    const files = [
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', '__init__.pyi'),
            content: 'def test(): ...',
        },
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', 'py.typed'),
            content: 'partial\n',
        },
        {
            path: combinePaths(typeshedFolder, 'stubs', 'myLibPackage', 'myLib.pyi'),
            content: '# empty',
        },
        {
            path: combinePaths(libraryRoot, 'myLib', '__init__.py'),
            content: 'def test(): pass',
        },
    ];

    // Stub packages win over typeshed.
    const importResult = getImportResult(files, ['myLib'], (c) => (c.typeshedPath = typeshedFolder));
    assert(importResult.isImportFound);
    assert(importResult.isStubFile);
    assert.strictEqual(
        1,
        importResult.resolvedPaths.filter((f) => f === combinePaths(libraryRoot, 'myLib', '__init__.pyi')).length
    );
});

test('typeshed fallback folder', () => {
    const files = [
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', '__init__.pyi'),
            content: 'def test(): ...',
        },
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', 'py.typed'),
            content: 'partial\n',
        },
        {
            path: combinePaths(typeshedFallback, 'stubs', 'myLibPackage', 'myLib.pyi'),
            content: '# empty',
        },
        {
            path: combinePaths(libraryRoot, 'myLib', '__init__.py'),
            content: 'def test(): pass',
        },
    ];

    // Stub packages win over typeshed.
    const importResult = getImportResult(files, ['myLib']);
    assert(importResult.isImportFound);
    assert(importResult.isStubFile);
    assert.strictEqual(
        1,
        importResult.resolvedPaths.filter((f) => f === combinePaths(libraryRoot, 'myLib', '__init__.pyi')).length
    );
});

test('py.typed file', () => {
    const files = [
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', '__init__.pyi'),
            content: 'def test(): ...',
        },
        {
            path: combinePaths(libraryRoot, 'myLib-stubs', 'py.typed'),
            content: 'partial\n',
        },
        {
            path: combinePaths(libraryRoot, 'myLib', '__init__.py'),
            content: 'def test(): pass',
        },
        {
            path: combinePaths(libraryRoot, 'myLib', 'py.typed'),
            content: '# typed',
        },
    ];

    // If py.typed file exists, then partial stub is disabled for this library.
    const importResult = getImportResult(files, ['myLib']);
    assert(importResult.isImportFound);
    assert(!importResult.isStubFile);
});

function getImportResult(
    files: { path: string; content: string }[],
    nameParts: string[],
    setup?: (c: ConfigOptions) => void
) {
    setup =
        setup ??
        ((c) => {
            /* empty */
        });

    const file = combinePaths('src', 'file.py');
    files.push({
        path: file,
        content: '# not used',
    });

    const fs = createFileSystem(files);
    const configOptions = getConfigOption(fs);
    setup(configOptions);

    const importResolver = new ImportResolver(fs, configOptions);
    const importResult = importResolver.resolveImport(file, configOptions.findExecEnvironment(file), {
        leadingDots: 0,
        nameParts: nameParts,
        importedSymbols: [],
    });

    return importResult;
}

function getConfigOption(fs: PyrightFileSystem) {
    const configOptions = new ConfigOptions(normalizeSlashes('/'));
    configOptions.venvPath = fs.getModulePath();
    configOptions.venv = fs.getModulePath();

    return configOptions;
}

function createFileSystem(files: { path: string; content: string }[]): PyrightFileSystem {
    const fs = new TestFileSystem(/* ignoreCase */ false, { cwd: normalizeSlashes('/') });

    for (const file of files) {
        const path = normalizeSlashes(file.path);
        const dir = getDirectoryPath(path);
        fs.mkdirpSync(dir);

        fs.writeFileSync(path, file.content);
    }

    return new PyrightFileSystem(fs);
}
